
#include "stdafx.h"
#include "CAuADXDecoder.h"

CAuADXDecoder::CAuADXDecoder()
{
  m_hFile = NULL;
  m_pPuc = NULL;
  m_pCs = new Nokomo::SyncObjects::CAuCriticalSection();
  m_nMasterVolume = DEFAULT_MASTER_VOLUME;
  m_szIniFile = NULL;
}

CAuADXDecoder::~CAuADXDecoder()
{
  Close();
  delete m_pPuc;
  delete m_pCs;
}

BOOL CAuADXDecoder::Open(LPSTR szFileName, SOUNDINFO* pInfo)
{
  Nokomo::SyncObjects::CAuCriticalSectionEx cs(m_pCs);

  m_hFile = CreateFile(szFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if(m_hFile)
  {
    if(ReadHeader())
    {
      CopyMemory(pInfo, &m_Info, sizeof(SOUNDINFO));
      Reset();
      return TRUE;
    }
    Close();
  }
  return FALSE;
}

// ̃t@CgȂƂĂԁDƂ delete 邾
void CAuADXDecoder::Close()
{
  CloseHandle(m_hFile);
  m_hFile = NULL;
}

DWORD CAuADXDecoder::Render(BYTE* pBuffer, DWORD dwSize)
{
  Nokomo::SyncObjects::CAuCriticalSectionEx cs(m_pCs);
  return DecodeBuffer(pBuffer, dwSize);
}

DWORD CAuADXDecoder::SetPosition(DWORD dwPos)
{
  Nokomo::SyncObjects::CAuCriticalSectionEx cs(m_pCs);
  DWORD dwPosSamples = dwPos / 1000 * m_Info.dwSamplesPerSec;

  if(m_Info.dwLoopFlag && dwPosSamples >= m_dwLoopStart)
  {
    SeekToLoopStartBlock();
    return (Nokomo::QWORD)m_dwLoopStart * 1000 / m_Info.dwSamplesPerSec;
  }
  else
  {
    Reset();
    return 0;
  }
}

VOID CAuADXDecoder::SeekToLoopStartBlock()
{
  // [vn_̃ubN̓ZN^ACĂ邾łȂ
  // [vn_̓ubN擪ɃACĂD
  DWORD dwBlockOffset = m_dwDataOffset + (m_dwLoopStart >> 5) * 18 * m_Info.dwChannels;
  Reset();
  SetFilePointer(m_hFile, dwBlockOffset, NULL, FILE_BEGIN);
  // ̓wb_ɓĂD0 ŏĂ̊Ȃ
  if(m_Header.byVersion == 3)
  {
    m_Context[0].s1 = m_Context[1].s1 = m_Header.v3.wLoopPointS1;
    m_Context[0].s2 = m_Context[1].s2 = m_Header.v3.wLoopPointS2;
  }
  else
  {
    m_Context[0].s1 = m_Context[1].s1 = m_Header.v4.wLoopPointS1;
    m_Context[0].s2 = m_Context[1].s2 = m_Header.v4.wLoopPointS2;
  }
  m_dwSamplesDecoded = m_dwLoopStart;
}


namespace
{
  // Big Endian 32bit  Little Endian DWORD 
  DWORD BEtoLE(DWORD dwBE)
  {
    // (;LDM)cc
    return ((dwBE & 0xff000000) >> 24)
         | ((dwBE & 0x00ff0000) >>  8)
         | ((dwBE & 0x0000ff00) <<  8)
         | ((dwBE & 0x000000ff) << 24)
         ;
  }

  WORD BEtoLE(WORD wBE)
  {
    return ((wBE & 0xff00) >> 8)
         | ((wBE & 0x00ff) << 8)
         ;
  }

  // BYTE* p  Big Endian  1word ǂށDǂ񂾌 p ͎ += 2 邱
  WORD ReadWordBE(BYTE* p) { return *p << 8 | *(p + 1); }

}

// #### private members

// m_hFile wb_ǂŉ
BOOL CAuADXDecoder::ReadHeader()
{
  DWORD dwBytesRead = 0;
  BOOL r;
  r = ReadFile(m_hFile, &m_Header, sizeof(SADXHeader), &dwBytesRead, NULL);
  if(!r || dwBytesRead != sizeof(SADXHeader)) return FALSE;

  // DWORD ȃõGfBAlXϊ
#define SWAP_DWORD_MEMBER(name) m_Header.name = BEtoLE(m_Header.name)
#define SWAP_WORD_MEMBER(name) m_Header.name = BEtoLE(m_Header.name)
  SWAP_DWORD_MEMBER(dwOffsetToSigAndData);
  SWAP_DWORD_MEMBER(dwSamplesPerSec);
  SWAP_DWORD_MEMBER(dwSamples);

  ZeroMemory(&m_Info, sizeof(SOUNDINFO));

  DWORD dwSignatureAddress;

  dwSignatureAddress = (m_Header.dwOffsetToSigAndData & 0x7fffffff) - 2;
  m_dwDataOffset = (m_Header.dwOffsetToSigAndData & 0x7fffffff) + 4;
  if(m_dwDataOffset >= GetFileSize(m_hFile, NULL)) return FALSE;

  switch(m_dwDataOffset)
  {
    // ̏Ƀ}b`ƁA[vʒuw肪 (c)CRI ɂ߂荞ł̂
    // I dwLoopFlag = FALSE
    case SIZEOF_ADX_V3:
      m_Header.byVersion = 3;
      break;
    case SIZEOF_ADX_V4:
      m_Header.byVersion = 4;
      break;
  }

  // ̕ӂ̕ςȒl KMP R
  m_Info.dwSamplesPerSec = m_Header.dwSamplesPerSec;
  m_Info.dwChannels = m_Header.byChannels;
  m_Info.dwBitsPerSample = 16; // TODO: 16bit ߂łˁH
  m_Info.dwUnitRender = (32 * m_Info.dwChannels * (m_Info.dwBitsPerSample >> 3)) * BLOCKS_PER_ONCE; // 1 ubN 32 Tv 16bit `l
  m_Info.dwMultiSongFlag = FALSE;
  m_Info.dwSeekable = TRUE;

  // TODO: ̃o[W莩̉cc
  switch(m_Header.byVersion)
  {
    case 3:
    {
      // TODO: ADX v3 ăt@CĂȂ́cc
      SWAP_DWORD_MEMBER(v3.dwLoopStart);
      SWAP_DWORD_MEMBER(v3.dwLoopEnd);
      SWAP_WORD_MEMBER(v3.wLoopPointS1);
      SWAP_WORD_MEMBER(v3.wLoopPointS2);
      m_dwLoopStart = m_Header.v3.dwLoopStart;
      m_dwLoopEnd = m_Header.v3.dwLoopEnd;
      break;
    }
    case 4:
    {
      SWAP_DWORD_MEMBER(v4.dwLoopStart);
      SWAP_DWORD_MEMBER(v4.dwLoopEnd);
      SWAP_WORD_MEMBER(v4.wLoopPointS1);
      SWAP_WORD_MEMBER(v4.wLoopPointS2);
      m_dwLoopStart = m_Header.v4.dwLoopStart;
      m_dwLoopEnd = m_Header.v4.dwLoopEnd;
      break;
    }
    default:
      return FALSE;
  }
#undef SWAP_DWORD_MEMBER
#undef SWAP_WORD_MEMBER
  // TODO: FIXME: [vL̔肪ρD
  // f[^ItZbg̒lwb_𒴂Ȃ
  // ܂pfBOȂƂƂȂ̂Ń[vȂD
  // ZN^pfBOKvȂ̂̓[vƂłD
  // o[WɊ֌WȂ炵
  m_dwSamples = m_Header.dwSamples;
  m_Info.dwLoopFlag = (m_dwDataOffset != (sizeof(SIZEOF_ADX_V3)) && m_dwDataOffset != (sizeof(SIZEOF_ADX_V4)))
                      && m_dwLoopStart < m_dwLoopEnd && m_dwLoopEnd <= m_dwSamples;
  if(m_szIniFile && GetPrivateProfileInt("kpiadx", "ForceNotToLoop", 0, m_szIniFile))
  {
    // [v𖳗IttOBfobOp
    m_Info.dwLoopFlag = 0;
  }

  int nLoopTimes = 2, nFadeoutOffset = 0;

  if(m_szIniFile)
  {
    nLoopTimes = GetPrivateProfileInt("kpiadx", "Loop", 2, m_szIniFile);
    nFadeoutOffset = GetPrivateProfileInt("kpiadx", "FadeoutOffset", 0, m_szIniFile);
  }

  if(m_Info.dwLoopFlag)
  {
    // TODO: FIXME: [v񌈂߂łH
    // TvPʂ͊ȒPɃI[o[t[̂ 64bit int
    Nokomo::QWORD qwSampleLength = m_dwLoopStart                         // Cgit@C擪烋[vJn܂Łj
                                 + (m_dwLoopEnd - m_dwLoopStart) * nLoopTimes;    // [v
    int nLength = qwSampleLength * 1000 / m_Info.dwSamplesPerSec + nFadeoutOffset;
    m_Info.dwLength = (nLength <= 0) ? 0 : nLength;
  }
  else
  {
    Nokomo::QWORD qwSampleLength = m_dwSamples;
    int nLength = qwSampleLength * 1000 / m_Info.dwSamplesPerSec + nFadeoutOffset;
    m_Info.dwLength = (nLength <= 0) ? 0 : nLength;
  }

  if(m_szIniFile)
    m_nMasterVolume = GetPrivateProfileInt("kpiadx", "MasterVolume", DEFAULT_MASTER_VOLUME, m_szIniFile);

  return TRUE;
}

// Open() Ct@C擪Đł悤
// V[NƂɂg
void CAuADXDecoder::Reset()
{
  SetFilePointer(m_hFile, m_dwDataOffset, NULL, FILE_BEGIN);
  m_dwSamplesDecoded = 0;
  ZeroMemory(m_Context, sizeof(SADXContext) * 2);
}

// 1w + 32h fR[h
void CAuADXDecoder::DecodeBlock(BYTE* pSource, short* pDest, SADXContext& context)
{
  int nScale = ReadWordBE(pSource);
  int d;
  int s0, s1 = context.s1, s2 = context.s2;
  int nMasterVolume = m_nMasterVolume;

  pSource += 2;
  for(int i = 0; i < 16; i++)
  {
    d = *pSource >> 4;
    if(d & 0x08)  d -= 0x10;
    s0 = (((d * nScale) * nMasterVolume) + ADXK0 * s1 - ADXK1 * s2) >> 14;
    if(s0 >  0x7fff)  s0 =  0x7fff;
    if(s0 < -0x7fff)  s0 = -0x7fff;
    *pDest++ = s0;  s2 = s1;  s1 = s0;

    d = *pSource & 0x0f;
    if(d & 0x08)  d -= 0x10;
    s0 = (((d * nScale) * nMasterVolume) + ADXK0 * s1 - ADXK1 * s2) >> 14;
    if(s0 >  0x7fff)  s0 =  0x7fff;
    if(s0 < -0x7fff)  s0 = -0x7fff;
    *pDest++ = s0;  s2 = s1;  s1 = s0;

    pSource++;
  }
  context.s1 = s1;
  context.s2 = s2;
}

// dwSize (% dwUnitRender == 0) ̕fR[h
DWORD CAuADXDecoder::DecodeBuffer(BYTE* pBuffer, DWORD dwSize)
{
  DWORD dwBytesDecoded = 0;
  DWORD dwBytesRead = 0;
  DWORD dwSamplesToCopy;
  BOOL r;
  DWORD dwBlocksToRead = BLOCKS_PER_ONCE * m_Info.dwChannels;
  BYTE* abySource = new BYTE[18 * dwBlocksToRead];

  if(m_Info.dwChannels == 1)
  {
    // m
    short sBuffer[32];

    while(dwBytesDecoded < dwSize)
    {
      r = ReadFile(m_hFile, abySource, 18 * dwBlocksToRead, &dwBytesRead, NULL);
      if(!r || dwBytesRead < 18) break;

      BYTE* pSource = abySource;
      while(pSource < abySource + dwBytesRead && dwBytesDecoded < dwSize)
      {
        DecodeBlock(pSource, sBuffer, m_Context[0]);
        pSource += 18;
        // Tv 32 samples ̔{Ƃ͌Ȃ̂ŁC
        // Rs[TCY̓TvŌ߂
        dwSamplesToCopy = (m_dwSamples - m_dwSamplesDecoded) > 32 ? 32 : m_dwSamples - m_dwSamplesDecoded;
        if(pBuffer)
          CopyMemory(pBuffer, sBuffer, dwSamplesToCopy * 2);

        pBuffer += dwSamplesToCopy * 2;
        dwBytesDecoded += dwSamplesToCopy * 2;
        m_dwSamplesDecoded += dwSamplesToCopy;
        if(dwSamplesToCopy != 32) goto BreakDecoding;
      }
    }
  }
  else
  {
    // XeI
    short sBufferL[32], sBufferR[32];

    while(dwBytesDecoded < dwSize)
    {
      // 
      r = ReadFile(m_hFile, abySource, 18 * dwBlocksToRead, &dwBytesRead, NULL);
      if(!r || dwBytesRead < 18) break;

      BYTE* pSource = abySource;
      while(pSource < abySource + dwBytesRead && dwBytesDecoded < dwSize)
      {
        DecodeBlock(pSource, sBufferL, m_Context[0]);
        pSource += 18;
        DecodeBlock(pSource, sBufferR, m_Context[1]);
        pSource += 18;

        // ADX ł͊e`l̃Tv 32 samples Ƃ
        // LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL RRRRRRRRRRRRRRRRRRRRRRRRRRRRRRRR ...
        // ƕł̂ŁC WaveAudio  LRLRLRLR ... `ɒ
        DWORD dwSamplesToCopy = (m_dwSamples - m_dwSamplesDecoded) > 32 ? 32 : m_dwSamples - m_dwSamplesDecoded;

        if(pBuffer)
        {
          short* pStereo = (short*)pBuffer;
          for(DWORD i = 0; i < dwSamplesToCopy; i++)
          {
            *pStereo++ = sBufferL[i];
            *pStereo++ = sBufferR[i];
          }
        }

        pBuffer += dwSamplesToCopy * 2 * 2;
        dwBytesDecoded += dwSamplesToCopy * 2 * 2;
        m_dwSamplesDecoded += dwSamplesToCopy;
        if(dwSamplesToCopy != 32) goto BreakDecoding;
      }
    }
  }
BreakDecoding:
  delete[] abySource;
  return dwBytesDecoded;
}
